import functools
import logging

from typing import Callable, Union, Optional, List

from ldap.ldapobject import SimpleLDAPObject as LdapObject
from ldap.filter import escape_filter_chars

from . import _ldaps, codecs, exceptions

logger = logging.getLogger(__name__)

UNSUPPORTED_SEARCH_ATTRIBUTES = ["thumbnailPhoto", "logonHours"]


def validate_query_params(query_params):
    unsupported_attributes = ", ".join([key for key in query_params.keys() if key in UNSUPPORTED_SEARCH_ATTRIBUTES])
    if unsupported_attributes:
        raise exceptions.ParameterError(f"Search contains unsupported attributes {unsupported_attributes}.")


def _build_query_params(query_params):
    # Takes a full dictionary of attribute -> value pairs and returns an LDAP filter.

    @functools.singledispatch
    def _escape(value, key):
        raise exceptions.ParameterError(
            "Only string, integer, and None values are supported in search functions. "
            f"Received search attribute {key} with value {value} which is type {type(value)}."
        )

    @_escape.register
    def _str(value: str, key):
        if value == "*":
            raise exceptions.ParameterError(
                f"Query attribute {key} has value {value}, which is not supported for performance reasons."
            )
        return (
            f"({key}={escape_filter_chars(value[:-1]) + '*'})"
            if value.endswith("*")
            else f"({key}={escape_filter_chars(value)})"
        )

    @_escape.register
    def _none(_: None, key):
        return f"(!({key}=*))"

    @_escape.register
    def _int(value: int, key):
        return f"({key}={value})"

    query_chunks = []
    for key, value in query_params.items():
        if any([isinstance(value, codecs.EncodedLdapBytes), isinstance(value, codecs.DecodedLdapString)]):
            raise exceptions.ParameterError(
                f"Attribute {key} has value {repr(value)}, which is not supported in LDAP queries. "
                "Please use a query encoder function to retrieve an EncodedQueryString object "
                "if this attribute requires a codec, or use a regular string as the value instead."
            )
        if any([isinstance(value, codecs.ProtectedString), isinstance(value, codecs.EncodedQueryString)]):
            # Neither a ProtectedValue or an EncodedQueryString should be escaped. These will either
            # have been generated by a query_encoder function in codecs.py, or created by the task author
            # for special attributes like objectGUID or objectSid.
            logger.debug(f"not escaping {value} as it is either a ProtectedString or an EncodedQueryString")
            query_chunks.append(f"({key}={value})")
        else:
            logger.debug(f"escaping value {value}")
            query_chunks.append(_escape(value, key))

    return "".join(query_chunks)


def get(
    query_base: str,
    connection: LdapObject,
    search_base: Optional[str] = None,
    search_scope: Optional[int] = None,
    **query_params,
) -> list:
    search_filter = _build_query_params(query_params)
    search_query = f"(&{query_base}{search_filter})" if search_filter else query_base
    return _ldaps._search(connection, search_query, search_base=search_base, search_scope=search_scope)  # type: ignore


def object_exists(_get: Callable, class_plural: str = "objects"):
    def _decorator(func):
        """ Decorates any function that assumes the target object
        already exists.

        This decorator also restricts functions so they can only
        operate against objects with the appropriate objectClass
        attribute. For example, an exception will be raised if you
        call users.delete with the distinguishedName of a group.

        The decorated function must take distinguished_name as the
        second positional argument.
        """

        @functools.wraps(func)
        def _fn(connection, distinguished_name: Union[List[str], str], *args, **kwargs):
            def _check_object_existence(dn):
                try:
                    _ = _get(connection, distinguishedName=dn)[0]
                except IndexError:
                    raise exceptions.NoSuchObject(f"No {class_plural} found with distinguishedName '{dn}'.")

            # Some functions (like users.not_member_of) will reference multiple
            # objects in a list, so we need to check the existence of all.
            if isinstance(distinguished_name, list):
                list(map(_check_object_existence, distinguished_name))
            else:
                _check_object_existence(distinguished_name)

            return func(connection, distinguished_name, *args, **kwargs)

        return _fn

    return _decorator


def modify(
    connection: LdapObject, distinguished_name: str, key: str, value: Union[int, str, bytes], operation: str
) -> None:
    ldap_operations = {"add": _ldaps._modify_add, "replace": _ldaps._modify_replace, "delete": _ldaps._modify_delete}

    if isinstance(value, int):
        value = str(value)
    if isinstance(value, str):
        value = bytes(value, "utf-8")

    try:
        ldap_operation = ldap_operations[operation]
    except KeyError:
        raise exceptions.ParameterError(
            f"Supplied operation {operation} not in supported list {list(ldap_operations.keys())}"
        )
    return ldap_operation(connection, distinguished_name, key, value)
