import logging
import functools

from typing import Callable, Any

from active_directory import client, exceptions, _object_classes, _ldaps

from .records import DomainController, User, UserData, Group, GroupData, Computer, ComputerData

logger = logging.getLogger(__name__)


def idempotent_create(func: Callable) -> Callable:
    @functools.wraps(func)
    def idempotent_create(*args, **kwargs) -> None:
        # Idempotent create functions should not care if the object already exists.
        try:
            return func(*args, **kwargs)
        except exceptions.ObjectExists as e:
            logger.debug("idempotent create function ignored the following exception")
            logger.exception(e)
            pass

    return idempotent_create


def idempotent_remove(func: Callable) -> Callable:
    @functools.wraps(func)
    def idempotent_remove(*args, **kwargs) -> None:
        # Idempotent remove functions should not care if the object does not exist.
        try:
            return func(*args, **kwargs)
        except exceptions.NoSuchObject as e:
            logger.debug("idempotent remove function ignored the following exception")
            logger.exception(e)
            pass

    return idempotent_remove


def ldap_function(func: Callable) -> Callable:
    @functools.wraps(func)
    def _fn(domain_controller: DomainController, *args, **kwargs) -> Any:
        # The opening of the connection should only happen if domain_controller isn't a SimpleLDAPObject yet.
        if not isinstance(domain_controller, _ldaps.ldap.ldapobject.LDAPObject):
            missing_keys = []
            for key in {"address", "domain", "username", "password"}:
                if key not in domain_controller:
                    missing_keys.append(key)
            if "port" not in domain_controller:
                domain_controller["port"] = 636
            if missing_keys:
                raise Exception(
                    "Active Directory functions require address, domain, username, and password in the "
                    f"connection details. Received {list(domain_controller.keys())}."
                )
            with client.initialize(**domain_controller)() as open_connection:
                return func(open_connection, *args, **kwargs)

        # Already a SimpleLDAPObject (decorated function calling another decorated functions).
        return func(domain_controller, *args, **kwargs)

    return _fn


def process_user(user: _object_classes.User) -> User:
    return UserData(user, inherited_properties=["is_disabled", "is_enabled", "is_expired", "is_locked", "password_age"]).__dict__


def process_group(group: _object_classes.Group) -> Group:
    return GroupData(group).__dict__


def process_computer(computer: _object_classes.Computer) -> Computer:
    return ComputerData(computer).__dict__
